#include "winlib_win32.c"

#include "unitabs.c"

#include "winlib_common.c"

void dump_event(ui_event_p ev)
{
	charmap_char_description_p cd;
	printf("=============================\nType: ");
	switch(ev->type)
	{
		case ui_key_event: puts("ui_key_event"); break;
		case ui_pointer_motion_event: puts("ui_pointer_motion_event"); break;
		case ui_window_focus_in_event: puts("ui_window_focus_in_event"); break;
		case ui_window_focus_out_event: puts("ui_window_focus_out_event"); break;
		case ui_window_expose_event: puts("ui_window_expose_event"); break;
		case ui_window_map_event: puts("ui_window_map_event"); break;
		case ui_window_unmap_event: puts("ui_window_unmap_event"); break;
		case ui_window_move_event: puts("ui_window_move_event"); break;
		case ui_window_size_event: puts("ui_window_size_event"); break;
		case ui_wm_event: printf("ui_wm_event (%d)\n",ev->configure.wm_event); break;
		default:
			puts("unknown");
			return;
	}
	printf("Time %g, x %d, y %d, rx %d, ry %d\n",
		ev->time, ev->client_x, ev->client_y, ev->screen_x, ev->screen_y);
	printf("Buton %x, state %x\n",ev->button,ev->state);
	printf("Press %u, char '%c', uni %x, scan %s\n",
		ev->key.press, (ev->key.ch >= 32) ? ev->key.ch : '?',
		ev->key.unicode,
		uiKeyscan2Name(ev->key.scan));
	if(ev->type == ui_key_event && ev->key.unicode)
	{
		cd = find_char_desc_by_uni(latin1_ccd,ev->key.unicode);
		if(!cd)
			cd = find_char_desc_by_uni(koi8_r_ccd,ev->key.unicode);
		if(cd)
			printf("\t%s\n",cd->char_name);
	}
	printf("Focused %u, visible %u\n",
		ev->configure.focused, ev->configure.visible);
	printf("Exposed: [%u,%u]-[%u,%u]\n",
		ev->configure.exposed_left,
		ev->configure.exposed_top,
		ev->configure.exposed_right,
		ev->configure.exposed_bottom);
	printf("Position [%d,%d]\n",
		ev->configure.x, ev->configure.y);
	printf("Size %u X %u\n",
		ev->configure.width, ev->configure.height);
}

int main(void)
{
	arch_window_p w;
	ui_event_t ev;

	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);

	w = archWinOpen("test",100,100,640,480,true,true);
	archWinShow(w,true);
	
	for(;;)
	{
		archWinPoll(w, true);
		while(w->msg_que)
		{
			DEQ_POP_BACK(event,w->msg_que,&ev);
			dump_event(&ev);
			if(ev.type != ui_key_event)
				continue;
			if(ev.key.scan == sc_escape)
				break;
			switch(ev.key.scan)
			{
				case sc_s:
					archWinShow(w, false);
					archWinPoll(w, false);
					while(w->msg_que)
					{
						DEQ_POP_BACK(event,w->msg_que,&ev);
						dump_event(&ev);
					}
					timeDelay(3000);
					archWinShow(w, true);
					break;
				case sc_p:
					archWinShowPointer(w, true);
					break;
				case sc_o:
					archWinShowPointer(w, false);
					break;
				case sc_w:
					archWinShowPointer(w, false);
					break;
				case sc_a:
					if(ev.key.press)
						archWinAutoRepeat(w,!w->enable_kb_repeat);
				default:
					break;
			}
		}
		if(ev.type == ui_key_event && ev.key.scan == sc_escape)
			break;
	}


	archWinClose(w);

	memPrintStats();
	return 0;
}

