#include <assert.h>

#include <stdio.h>
#include <string.h>
#ifdef	__BSD__
#include <stdlib.h>
#else
#include <malloc.h>
#endif	/* __BSD__ */

#include "ui_event.h"
#include "ui_event.c"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <GL/gl.h>
#include <GL/glx.h>
#ifdef	__BSD__
#include <unistd.h> /* for select() */
#else
#include <sys/select.h>
#endif	/* __BSD__ */
#include <sys/time.h>
#include <sys/types.h>

#define	HAVE_XF86_VIDMODE
/*
 * Uncomment to use XFree86 VidMode extension
 */

#ifdef	HAVE_XF86_VIDMODE
#include <X11/extensions/xf86vmode.h>
#endif	/* HAVE_XF86_VIDMODE */

Display 	*dpy;
int 		scrnum;
Window		root, wnd;



/* GLOBS */

XIM			wnd_im;
XIC			wnd_ic;

Atom		wnd_delete;
unsigned	wnd_event_mask;
unsigned	wnd_start_time = 0;
int 		wnd_focused = 0;
int			wnd_visible = 0;
int			wnd_width, wnd_height;
int			wnd_x, wnd_y;
unsigned	wnd_last_key_state = 0;
unsigned	wnd_last_button_state = 0;
unsigned	wnd_keycode_to_scan[256];
unsigned char wnd_kb_state[sc_maxcode];

XSetWindowAttributes attr;

#ifdef	HAVE_XF86_VIDMODE

XF86VidModeModeInfo **vidmodes = NULL;
int num_vidmodes = 0;
int vidmode_changed = 0;

int
X11_set_video_mode(int width, int height)
{
	int mj = 0, mn = 0;
	int sw, sh, i;
	sw = WidthOfScreen(DefaultScreenOfDisplay(dpy));
	sh = HeightOfScreen(DefaultScreenOfDisplay(dpy));
	if (width == sw && height == sh)
		return !0;
	if (!vidmodes) {
		if (!XF86VidModeQueryVersion(dpy, &mj, &mn)) {
			fprintf(stderr, "XFree86-VidModeExtension not found.\n");
			return 0;
		}
		printf("found XFree86-VidModeExtension (version %d.%d).\n", mj, mn);
		XF86VidModeGetAllModeLines(dpy, scrnum, &num_vidmodes, &vidmodes);
		if (!vidmodes) { /* just a safety measure, though... */
			fprintf(stderr, "XF86VidModeGetAllModeLines() failed !.\n");
			return 0;
		}
	}
	for (i = 0; i < num_vidmodes; i++) {
		if (width == vidmodes[i]->hdisplay &&
			height == vidmodes[i]->vdisplay)
			break;
	}
	if (i == num_vidmodes) {
		fprintf(stderr, "Unable to find mode '%i X %i'.\n", width, height);
		return 0;
	}
	XWarpPointer(dpy, None, wnd, 0, 0, 0, 0, width >> 1, height >> 1);
	XF86VidModeSwitchToMode(dpy, scrnum, vidmodes[i]);
	XF86VidModeSetViewPort(dpy, scrnum, 0, 0);
	XF86VidModeLockModeSwitch(dpy, scrnum, True);
	/* TODO: viewport locking (but _how_? - have to look to Q code ?) */
	vidmode_changed = 1;
	return !0;
}

void
X11_restore_video_mode(void)
{
	int i;
	if (!vidmodes)
		return;
	if (vidmode_changed)
	{
		XF86VidModeLockModeSwitch(dpy, scrnum, False);
		XF86VidModeSwitchToMode(dpy, scrnum, vidmodes[0]);
		vidmode_changed = 0;
	}
	for (i = 0; i < num_vidmodes; i++) {
		if (vidmodes[0]->privsize)
#if	defined(__cplusplus) || defined(c_plusplus)
			XFree(vidmodes[0]->c_private);
#else
			XFree(vidmodes[0]->private);
#endif
	}
	XFree(vidmodes);
	vidmodes = NULL;
}

#else	/* !HAVE_XF86_VIDMODE */

int
X11_set_video_mode(int width, int height)
{
	return 0;
}

void
X11_restore_video_mode(void) 
{
}

#endif	/* !HAVE_XF86_VIDMODE */

/*
 * ref_shutdown()
 */
void
ref_shutdown(void)
{
	if (wnd)
	{
		XDestroyWindow(dpy, wnd);
		wnd = 0;
	}
	X11_restore_video_mode();
	if (dpy)
	{
		XCloseDisplay(dpy);
		dpy = NULL;
	}
	/*
	XDestroyIC(wnd_ic);
	XCloseIM(wnd_im);
	*/
}

#include <stdlib.h>
#include <locale.h>

static void
X11_init_keycode_to_scan(Display *dpy,unsigned *tab);

/*
 * ref_init()
 */
int
ref_init(char *caption, int width, int height, int fullscreen)
{
	unsigned long mask;
	XSizeHints sh;
	dpy = XOpenDisplay(NULL);
	if (!dpy) {
		fprintf(stderr, "XOpenDisplay() failed.\n");
		return 0;
	}

	X11_init_keycode_to_scan(dpy, wnd_keycode_to_scan);

	scrnum = DefaultScreen(dpy);
	root = RootWindow(dpy, scrnum);

	{
		/* -> LibVCreateRoot() */
		char *lang = getenv("LANG");
		char *deflang = "en_US";
		if(!lang)
		{
			
			fprintf(stderr, "warning: LANG not set, assuming \"%s\".\n",deflang);
			lang = deflang;
		}
		if(!setlocale(LC_ALL,lang))
		{
			fprintf(stderr, "warning: could not set locale \"%s\"\n", lang);
		}
		else
			printf("Using locale \"%s\"\n", lang);
	}

	attr.border_pixel = None;
	wnd_event_mask = attr.event_mask = 
		StructureNotifyMask |
		ExposureMask |
		KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ButtonMotionMask | VisibilityChangeMask | FocusChangeMask;
	attr.do_not_propagate_mask = KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ButtonMotionMask;
	attr.background_pixel = WhitePixel(dpy, scrnum);
	attr.backing_store = NotUseful;
	attr.save_under = False;
	mask = CWBorderPixel | CWEventMask | CWDontPropagate |
		CWBackPixel | CWBackingStore | CWSaveUnder;
	if (fullscreen) {
		mask |= CWOverrideRedirect;
		attr.override_redirect = True;
	}

	wnd_x = 0; /* fixme */
	wnd_y = 0;

	if (width == 0)
		width = DisplayWidth(dpy, scrnum) - wnd_x;
	if (height == 0)
		height = DisplayHeight(dpy, scrnum) - wnd_y;

	wnd_width = width;
	wnd_height = height;

	wnd = XCreateWindow(dpy, root, wnd_x, wnd_y, width, height, 0,
						DefaultDepth(dpy,scrnum), InputOutput,
						DefaultVisual(dpy,scrnum),
						mask, &attr);
	if (!wnd) {
		fprintf(stderr, "XCreateWindow() failed.\n");
		ref_shutdown();
		return 0;
	}

	XSetLocaleModifiers("");
	wnd_im = XOpenIM(dpy,NULL,NULL,NULL);
	wnd_ic = XCreateIC(wnd_im,XNInputStyle,
		XIMPreeditNothing | XIMStatusNothing,
		XNClientWindow,wnd,XNFocusWindow,wnd,NULL);

	XStoreName(dpy, wnd, caption);
	memset(&sh, 0, sizeof(sh));
	/*
	  sh.flags = USPosition | USSize | PPosition | PSize | \
	  PMinSize | PMaxSize | PBaseSize;
	*/
	sh.flags = USPosition | PBaseSize;
	sh.x = 0;
	sh.y = 0;
	sh.width = sh.min_width = sh.max_width = sh.base_width = width;
	sh.height = sh.min_height = sh.max_height = sh.base_height = height;
	XSetWMNormalHints(dpy, wnd, &sh);
	wnd_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(dpy, wnd, &wnd_delete, 1);
	XMapRaised(dpy, wnd);
	if (fullscreen)
		XSetInputFocus(dpy, wnd, RevertToParent, CurrentTime);
	return !0;
}

static void
X11_init_keycode_to_scan(Display *dpy,unsigned *tab)
{
	KeySym ks;
	unsigned scan;
	int i, mincode, maxcode;
	XDisplayKeycodes(dpy, &mincode, &maxcode);
	if(mincode < 0)
		mincode = 0;
	if(maxcode > 255)
		maxcode = 255;
	for(i = 0;i < mincode;i++)
		tab[i] = sc_unknown;
	for(i = maxcode + 1;i < 256;i++)
		tab[i] = sc_unknown;
	for(i = mincode;i <= maxcode;i++)
	{
		ks = XKeycodeToKeysym(dpy,i,0);
		switch(ks)
		{
            case XK_KP_Page_Up: 
            case XK_KP_9: scan = sc_kp_prior; break;
            case XK_Page_Up: scan = sc_prior; break;

            case XK_KP_Page_Down: 
            case XK_KP_3: scan = sc_kp_next; break;
            case XK_Page_Down: scan = sc_next; break;

            case XK_KP_Home:
            case XK_KP_7: scan = sc_kp_home; break;
            case XK_Home: scan = sc_home; break;

            case XK_KP_End:
            case XK_KP_1: scan = sc_kp_end; break;
            case XK_End: scan = sc_end; break;

            case XK_KP_Left:
            case XK_KP_4: scan = sc_kp_left; break;
            case XK_Left: scan = sc_left; break;

            case XK_KP_Right:
            case XK_KP_6: scan = sc_kp_right; break;
            case XK_Right: scan = sc_right; break;

            case XK_KP_Down:
            case XK_KP_2: scan = sc_kp_down; break;
            case XK_Down: scan = sc_down; break;

            case XK_KP_Up:   
            case XK_KP_8: scan = sc_kp_up; break;
            case XK_Up: scan = sc_up; break;

            case XK_Escape: scan = sc_escape; break;

            case XK_KP_Enter: scan = sc_kp_enter; break;
            case XK_Return: scan = sc_enter; break;

            case XK_Tab: scan = sc_tab; break;

            case XK_F1: scan = sc_f1; break;
            case XK_F2: scan = sc_f2; break;
            case XK_F3: scan = sc_f3; break;
            case XK_F4: scan = sc_f4; break;
            case XK_F5: scan = sc_f5; break;
            case XK_F6: scan = sc_f6; break;
            case XK_F7: scan = sc_f7; break;
            case XK_F8: scan = sc_f8; break;
            case XK_F9: scan = sc_f9; break;
            case XK_F10: scan = sc_f10; break;
            case XK_F11: scan = sc_f11; break;
            case XK_F12: scan = sc_f12; break;

            case XK_BackSpace: scan = sc_backspace; break;
            case XK_KP_Delete:
            case XK_KP_Decimal: scan = sc_kp_delete; break;
            case XK_Delete: scan = sc_delete; break;

            case XK_Pause: scan = sc_pause; break;

            case XK_Shift_L: scan = sc_lshift; break;
            case XK_Shift_R: scan = sc_rshift; break;

            case XK_Execute:
            case XK_Control_L: scan = sc_lcontrol; break;
            case XK_Control_R: scan = sc_rcontrol; break;

            case XK_Alt_L:  
            case XK_Meta_L: scan = sc_lalt; break;
            case XK_Alt_R:  
            case XK_Meta_R: scan = sc_ralt; break;

            case XK_KP_Begin: scan = sc_kp_begin; break;

            case XK_KP_Insert:
            case XK_KP_0: scan = sc_kp_insert; break;
            case XK_Insert: scan = sc_insert; break;

            case XK_KP_Multiply: scan = sc_kp_multiply; break;
            case XK_KP_Add: scan = sc_kp_add; break;
            case XK_KP_Subtract: scan = sc_kp_subtract; break;
            case XK_KP_Divide: scan = sc_kp_slash; break;

            case XK_exclam: scan = sc_1; break;
            case XK_at: scan = sc_2; break;
            case XK_numbersign: scan = sc_3; break;
            case XK_dollar: scan = sc_4; break;
            case XK_percent: scan = sc_5; break;
            case XK_asciicircum: scan = sc_6; break;
            case XK_ampersand: scan = sc_7; break;
            case XK_asterisk: scan = sc_8; break;
            case XK_parenleft: scan = sc_9; break;
            case XK_parenright: scan = sc_0; break;
            case XK_twosuperior: scan = sc_grave; break;
            case XK_space:
            case XK_KP_Space: scan = sc_space; break;
		
		    case XK_Sys_Req:
		    case XK_Print: scan = sc_print; break;
		    case XK_Scroll_Lock: scan = sc_scroll_lock; break;
		    case XK_Caps_Lock: scan = sc_caps_lock; break;
		    case XK_Num_Lock: scan = sc_num_lock; break;
		    case XK_Hyper_L:
		    case XK_Super_L: scan = sc_lsuper; break;
		    case XK_Hyper_R:
		    case XK_Super_R: scan = sc_rsuper; break;
		    case XK_Menu: scan = sc_menu; break;

		    case '`':
		    case '~': scan = sc_grave; break;
		    case '1': scan = sc_1; break;
		    case '2': scan = sc_2; break;
		    case '3': scan = sc_3; break;
		    case '4': scan = sc_4; break;
		    case '5': scan = sc_5; break;
		    case '6': scan = sc_6; break;
		    case '7': scan = sc_7; break;
		    case '8': scan = sc_8; break;
		    case '9': scan = sc_9; break;
		    case '0': scan = sc_0; break;
		    case '-':
		    case '_': scan = sc_minus; break;
		    case '=':
		    case '+': scan = sc_equal; break;
		    case '\\':
		    case '|': scan = sc_backslash; break;
		    case 'q':
		    case 'Q': scan = sc_q; break;
		    case 'w':
		    case 'W': scan = sc_w; break;
		    case 'e':
		    case 'E': scan = sc_e; break;
		    case 'r':
		    case 'R': scan = sc_r; break;
		    case 't':
		    case 'T': scan = sc_t; break;
		    case 'y':
		    case 'Y': scan = sc_y; break;
		    case 'u':
		    case 'U': scan = sc_u; break;
		    case 'i':
		    case 'I': scan = sc_i; break;
		    case 'o':
		    case 'O': scan = sc_o; break;
		    case 'p':
		    case 'P': scan = sc_p; break;
		    case '[':
		    case '{': scan = sc_lbracket; break;
		    case ']':
		    case '}': scan = sc_rbracket; break;
		    case 'a':
		    case 'A': scan = sc_a; break;
		    case 's':
		    case 'S': scan = sc_s; break;
		    case 'd':
		    case 'D': scan = sc_d; break;
		    case 'f':
		    case 'F': scan = sc_f; break;
		    case 'g':
		    case 'G': scan = sc_g; break;
		    case 'h':
		    case 'H': scan = sc_h; break;
		    case 'j':
		    case 'J': scan = sc_j; break;
		    case 'k':
		    case 'K': scan = sc_k; break;
		    case 'l':
		    case 'L': scan = sc_l; break;
		    case ';':
		    case ':': scan = sc_semicolon; break;
		    case '\'':
		    case '"': scan = sc_apostrophe; break;
		    case 'z':
		    case 'Z': scan = sc_z; break;
		    case 'x':
		    case 'X': scan = sc_x; break;
		    case 'c':
		    case 'C': scan = sc_c; break;
		    case 'v':
		    case 'V': scan = sc_v; break;
		    case 'b':
		    case 'B': scan = sc_b; break;
		    case 'n':
		    case 'N': scan = sc_n; break;
		    case 'm':
		    case 'M': scan = sc_m; break;
		    case ',':
		    case '<': scan = sc_comma; break;
		    case '.':
		    case '>': scan = sc_period; break;
		    case '/':
		    case '?': scan = sc_slash; break;
			default: scan = sc_unknown;
		}
		tab[i] = scan;
	}
}

static void
X11_make_key_codes(XEvent *event,ui_key_event_p dst)
{
	int nb, nub = 0;
	unsigned char buf[260];
	wchar_t wbuf[260];
	KeySym ks;
	int saved_type;
	/*
	 * This is a hack to make XwcLookupString() work for KeyRelease
	 * event as for KeyPress.
	 */
	saved_type = event->type;
	event->type = KeyPress;
	nub = XwcLookupString(wnd_ic, &event->xkey, wbuf, 256, &ks, NULL);
	nb = XLookupString(&event->xkey, (char*)buf, 256, &ks, NULL);
	event->type = saved_type;
	/* now forget about the new keysym ... */
	dst->scan = wnd_keycode_to_scan[event->xkey.keycode];
	dst->ch = (nb > 0) ? buf[0] : 0;
	dst->unicode = (nub > 0) ? wbuf[0] : 0;
}

void
X11_grab_pointer(int grab)
{
	static int pointer_grabbed = 0;
	if(grab)
	{
		if (pointer_grabbed)
			return;
		XGrabPointer(dpy, wnd, True, 0, GrabModeAsync, GrabModeAsync, wnd,
			None, CurrentTime);
		pointer_grabbed = 1;
	}
	else
	{
		if (!pointer_grabbed)
			return;
		pointer_grabbed = 0;
		XUngrabPointer(dpy, CurrentTime);
	}
}

static void
X11_make_key_states(unsigned x_state, ui_event_p ev)
{
	ev->state = ev->button = 0;

	if(x_state & Mod2Mask)
		ev->state |= UI_NUM_LOCK_MASK;
	if(x_state & LockMask)
		ev->state |= UI_CAPS_LOCK_MASK;
	if(x_state & Mod5Mask)
		ev->state |= UI_SCROLL_LOCK_MASK;
	if(x_state & ShiftMask)
		ev->state |= UI_SHIFT_MASK;
	if(x_state & ControlMask)
		ev->state |= UI_CONTROL_MASK;
	if(x_state & Mod1Mask)
		ev->state |= UI_ALT_MASK;
	if(x_state & Mod4Mask)
		ev->state |= UI_WIN_MASK;

	if(x_state & Button1Mask)
		ev->button |= UI_LBUTTON_MASK;
	if(x_state & Button2Mask)
		ev->button |= UI_MBUTTON_MASK;
	if(x_state & Button3Mask)
		ev->button |= UI_RBUTTON_MASK;

	wnd_last_key_state = ev->state;
	wnd_last_button_state = ev->button;
}

static void
make_default_key_event(ui_key_event_p ev)
{
	ev->press = false;
	ev->scan = sc_unknown;
	ev->ch = 0;
	ev->unicode = 0;
}

static void
set_configure_from_globs(ui_configure_event_p ev)
{
	ev->focused = wnd_focused;
	ev->visible = wnd_visible;
	ev->x = wnd_x;
	ev->y = wnd_y;
	ev->width = wnd_width;
	ev->height = wnd_height;
	ev->exposed_left = ev->exposed_top = 0;
	ev->exposed_right = ev->exposed_bottom = 0;
}

static vec_t
X11_make_time(unsigned *start,unsigned current)
{
	Time delta1, delta2;
	if(!*start)
		*start = current; /* FIXME: from what to count the time ??? */
	if(*start <= current)
	{
		delta1 = current - *start;
		delta2 = 0;
	}
	else
	{
		/* overflow ! */
		unsigned last;
#ifdef _ARCH_64_
		last = 0xFFFFFFFFFFFFFFFF;
#else
		last = 0xFFFFFFFF;
#endif
		delta1 = last - *start;
		delta2 = current;
		/* FIXME: this is wrong */
		*start = current;
	}
	return (vec_t)0.001 * ((vec_t)delta1 + (vec_t)delta2);
}

static ui_event_type_t
X11_process_message(XEvent *event,ui_event_p ev_ret)
{
	static ui_event_t stub;
	ui_event_p dst = ev_ret ? ev_ret : &stub;
	switch(event->type)
	{
	case KeyPress:
	case KeyRelease:

		dst->type = ui_key_event;
		dst->time = X11_make_time(&wnd_start_time, event->xkey.time);
		dst->client_x = event->xkey.x;
		dst->client_y = event->xkey.y;
		dst->screen_x = event->xkey.x_root;
		dst->screen_y = event->xkey.y_root;
		dst->key.press = (event->type == KeyPress) ? true : false;
		X11_make_key_states(event->xkey.state, dst);
		X11_make_key_codes(event,&dst->key);
		set_configure_from_globs(&dst->configure);
		if(dst->key.scan != sc_unknown)
			wnd_kb_state[dst->key.scan] =
				(event->type == KeyPress) ? 1 : 0;
		break;

	case ButtonPress:
	case ButtonRelease:
	
		dst->type = ui_key_event;
		dst->time = X11_make_time(&wnd_start_time, event->xbutton.time);
		dst->client_x = event->xbutton.x;
		dst->client_y = event->xbutton.y;
		dst->screen_x = event->xbutton.x_root;
		dst->screen_y = event->xbutton.y_root;
		X11_make_key_states(event->xbutton.state, dst);
		dst->key.press = (event->type == ButtonPress) ? true : false;
		switch(event->xbutton.button)
		{
			case Button1: dst->key.scan = sc_lbutton; break;
			case Button2: dst->key.scan = sc_mbutton; break;
			case Button3: dst->key.scan = sc_rbutton; break;
			default: dst->key.scan = sc_unknown;
		}
		dst->key.ch = dst->key.unicode = 0;
		set_configure_from_globs(&dst->configure);
		if(dst->key.scan != sc_unknown)
			wnd_kb_state[dst->key.scan] =
				(event->type == ButtonPress) ? 1 : 0;
		break;

	case MotionNotify:

		dst->type = ui_pointer_motion_event;
		dst->time = X11_make_time(&wnd_start_time, event->xmotion.time);
		dst->client_x = event->xmotion.x;
		dst->client_y = event->xmotion.y;
		dst->screen_x = event->xmotion.x_root;
		dst->screen_y = event->xmotion.y_root;
		X11_make_key_states(event->xkey.state, dst);
		make_default_key_event(&dst->key);
		set_configure_from_globs(&dst->configure);
		break;

	case FocusIn:
	case FocusOut:
		
		if(event->type == FocusIn)
		{
			wnd_focused = true;
			dst->type = ui_window_focus_in_event;
		}
		else
		{
			wnd_focused = false;
			dst->type = ui_window_focus_out_event;
		}
		dst->time = 0;
		dst->client_x = 0;
		dst->client_y = 0;
		dst->screen_x = 0;
		dst->screen_y = 0;
		make_default_key_event(&dst->key);
		set_configure_from_globs(&dst->configure);
		dst->state = wnd_last_key_state;
		dst->button = wnd_last_button_state;
		break;

	case ConfigureNotify:
		
		if(event->xconfigure.width != wnd_width ||
				event->xconfigure.height != wnd_height)
			dst->type = ui_window_size_event;
		else
			dst->type = ui_window_move_event;
		dst->time = 0;
		dst->client_x = 0;
		dst->client_y = 0;
		dst->screen_x = 0;
		dst->screen_y = 0;
		wnd_x = event->xconfigure.x;
		wnd_y = event->xconfigure.y;
		wnd_width = event->xconfigure.width;
		wnd_height = event->xconfigure.height;
		make_default_key_event(&dst->key);
		set_configure_from_globs(&dst->configure);
		dst->state = wnd_last_key_state;
		dst->button = wnd_last_button_state;
		break;

	case MapNotify:
	case UnmapNotify:
		
		if(event->type == MapNotify)
		{
			wnd_visible = true;
			dst->type = ui_window_map_event;
		}
		else
		{
			wnd_visible = false;
			dst->type = ui_window_unmap_event;
		}
		dst->time = 0;
		dst->client_x = 0;
		dst->client_y = 0;
		dst->screen_x = 0;
		dst->screen_y = 0;
		make_default_key_event(&dst->key);
		set_configure_from_globs(&dst->configure);
		dst->state = wnd_last_key_state;
		dst->button = wnd_last_button_state;
		break;

	case ClientMessage:
		if((Atom)event->xclient.data.l[0] != wnd_delete)
			return ui_no_event;
		dst->type = ui_wm_event;
		dst->time = 0;
		dst->client_x = 0;
		dst->client_y = 0;
		dst->screen_x = 0;
		dst->screen_y = 0;
		make_default_key_event(&dst->key);
		set_configure_from_globs(&dst->configure);
		dst->state = wnd_last_key_state;
		dst->button = wnd_last_button_state;
		dst->configure.wm_event = wm_close;
		break;

	default:

		return ui_no_event;
	}

	return dst->type;
}

ui_event_type_t
ref_peek_event(ui_event_p ev_ret)
{
	XEvent event;
	int l, t, r, b, left, top, right, bottom;
	/*
	 * Remove all Expose events from queue and collapse them
	 */
	left = top = 0x7FFFFFFF;
	right = bottom = -1;
	while(XCheckTypedWindowEvent(dpy,wnd,Expose,&event))
	{
		l = event.xexpose.x;
		t = event.xexpose.y;
		r = l + event.xexpose.width;
		b = t + event.xexpose.height;
		if(left > l)
			left = l;
		if(top > t)
			top = t;
		if(right < r)
			right = r;
		if(bottom < b)
			bottom = b;
	}
	if(right != -1)
	{
		ev_ret->type = ui_window_expose_event;
		ev_ret->time = 0;
		ev_ret->client_x = 0;
		ev_ret->client_y = 0;
		ev_ret->screen_x = 0;
		ev_ret->screen_y = 0;
		make_default_key_event(&ev_ret->key);
		set_configure_from_globs(&ev_ret->configure);
		ev_ret->configure.exposed_left = left;
		ev_ret->configure.exposed_top = top;
		ev_ret->configure.exposed_right = right;
		ev_ret->configure.exposed_bottom = bottom;
		ev_ret->state = wnd_last_key_state;
		ev_ret->button = wnd_last_button_state;
		return ev_ret->type;
	}
	if(XCheckWindowEvent(dpy,wnd,wnd_event_mask,&event))
		return X11_process_message(&event, ev_ret);
	return ui_no_event;
}

void
ref_delay(int ms)
{
	/*
	 * The straightforward way is to use usleep(), but there are some
	 * bizzare problems with headers (even) on modern Linux
	 * distrubutions.
	 * And: this approach also works under FreeBSD.
     * The idea is _not_ mine; actually, it's quite old...
	 */
	struct timeval	delay;
	delay.tv_sec = ms / 1000;
	delay.tv_usec = 1000 * (ms % 1000);
	select(0, NULL, NULL, NULL, &delay);
}

#include "unitabs.c"

void dump_event(ui_event_p ev)
{
	charmap_char_description_p cd;
	printf("=============================\nType: ");
	switch(ev->type)
	{
		case ui_key_event: puts("ui_key_event"); break;
		case ui_pointer_motion_event: puts("ui_pointer_motion_event"); break;
		case ui_window_focus_in_event: puts("ui_window_focus_in_event"); break;
		case ui_window_focus_out_event: puts("ui_window_focus_out_event"); break;
		case ui_window_expose_event: puts("ui_window_expose_event"); break;
		case ui_window_map_event: puts("ui_window_map_event"); break;
		case ui_window_unmap_event: puts("ui_window_unmap_event"); break;
		case ui_window_move_event: puts("ui_window_move_event"); break;
		case ui_window_size_event: puts("ui_window_size_event"); break;
		default:
			puts("unknown");
			return;
	}
	printf("Time %g, x %d, y %d, rx %d, ry %d\n",
		ev->time, ev->client_x, ev->client_y, ev->screen_x, ev->screen_y);
	printf("Buton %x, state %x\n",ev->button,ev->state);
	printf("Press %u, char '%c', uni %x, scan %s\n",
		ev->key.press, (ev->key.ch >= 32) ? ev->key.ch : '?',
		ev->key.unicode,
		uiKeyscan2Name(ev->key.scan));
	if(ev->type == ui_key_event && ev->key.unicode)
	{
		cd = find_char_desc_by_uni(latin1_ccd,ev->key.unicode);
		if(!cd)
			cd = find_char_desc_by_uni(koi8_r_ccd,ev->key.unicode);
		if(cd)
			printf("\t%s\n",cd->char_name);
	}
	printf("Focused %u, visible %u\n",
		ev->configure.focused, ev->configure.visible);
	printf("Exposed: [%u,%u]-[%u,%u]\n",
		ev->configure.exposed_left,
		ev->configure.exposed_top,
		ev->configure.exposed_right,
		ev->configure.exposed_bottom);
	printf("Position [%d,%d]\n",
		/*ev->configure.x, ev->configure.y*/wnd_x, wnd_y);
	printf("Size %u X %u\n",
		ev->configure.width, ev->configure.height);
}

static void
X11_get_keyboard_state(Display *dpy,unsigned char *state_ret)
{
	unsigned keycode, i, idx;
	unsigned char c, mask, keys[32], map[256];
	XQueryKeymap(dpy,(char*)keys);
	for(i = idx = 0;idx < 32;idx++)
	{
		c = keys[idx];
		mask = 1;
		while(mask)
		{
			map[i++] = (c & mask) ? 1 : 0;
			mask <<= 1;
		}
	}
	memset(state_ret,0,sc_maxcode);
	for(keycode = 0;keycode < 256;keycode++)
		state_ret[wnd_keycode_to_scan[keycode]] = map[keycode];
}

unsigned char last_state[256];

int main(void)
{
	ui_event_t ev;
	if(!ref_init("demo1", 640, 480, 0))
		return -1;
	do
	{
		if(ref_peek_event(&ev) == ui_no_event)
		{
			if(!wnd_focused)
			{
				unsigned scan;
				unsigned char state[sc_maxcode];
				X11_get_keyboard_state(dpy,state);
				for(scan = 1; scan < sc_maxcode;scan++)
					if(last_state[scan] != state[scan])
					{
						printf("Scan '%s': %d->%d\n",
							uiKeyscan2Name(scan), last_state[scan], state[scan]);
						last_state[scan] = state[scan];
					}
			}
			ref_delay(20);
			continue;
		}
		dump_event(&ev);
	} while (ev.key.scan != sc_escape);
	assert(XSetLocaleModifiers(NULL));
	ref_shutdown();
	return 0;
}
